/* ----
 * ---- file   : mlb.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel.
 * ----          Distributed under terms of the MIT LICENSE (MIT).
 * ----
 * ---- Permission is hereby granted, free of charge, to any person obtaining a copy
 * ---- of this software and associated documentation files (the "Software"), to deal
 * ---- in the Software without restriction, including without limitation the rights
 * ---- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * ---- copies of the Software, and to permit persons to whom the Software is
 * ---- furnished to do so, subject to the following conditions:
 * ----
 * ---- The above copyright notice and this permission notice shall be included in
 * ---- all copies or substantial portions of the Software.
 * ----
 * ---- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * ---- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * ---- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * ---- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * ---- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * ---- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * ---- THE SOFTWARE.
 * ----
 * ---- info   : This is part of the "c64_tools" package.
 * ----
 * ---- changed: 10Sep2013, 11Sep2013, 12Sep2013, 13Sep2013, 14Sep2013, 15Sep2013
 * ----
 * ----
 */

#include <std.h>
#include <log.h>
#include <mem.h>
#include <sem.h>
#include <sys.h>
#include <tsk.h>
#include <hwi.h>
#include <bcache.h>

#include <c64.h>

#include <string.h>  /* for memset() */

////#include <csl.h>
////#include <cslr.h>
//////#include <soc.h>
////#include <csl_intc.h>
////#include <csl_intcAux.h>

#include "../../../include/types.h"
#include "../../../include/arch/dm3730/dm3730.h"
#include "../../../include/dsp_common.h"
#include "../../../include/dsp_config.h"

#include "../include/int.h"
#include "../include/mlb.h"


#define MLB_INT_VEC_ID  IVA2_INT_10


#define MLB_SRAM_SIZE  (32u)
#pragma DATA_ALIGN(mlb_sram,  32);
#pragma DATA_SECTION(mlb_sram,  ".sec_mlb_sram");
unsigned char mlb_sram[MLB_SRAM_SIZE];


/* --------------------------------------------------------------------------- vars */
static TSK_Attrs mlb_task_attrs;
static TSK_Handle mlb_task_handle;

//static TSK_Attrs watchdog_task_attrs;
//static TSK_Handle watchdog_task_handle;

//static SEM_Handle mlb_sem_handle;

static volatile sUI num_irqs;
static volatile sUI num_msg_recvd;

static volatile sU32 mlb_inbox[4];

static dsp_component_t components[MLB_MAX_COMPONENTS];
static sUI num_components;


/* --------------------------------------------------------------------------- loc_component_find_by_id() */
dsp_component_t *loc_component_find_by_id(dsp_component_id_t _id) {
   sUI i;
   dsp_component_t *ret = components;

   /* (todo) optimize */
   for(i=0; i<num_components; i++)
   {
      if(ret->component_id == _id)
      {
         /* Found component */
         break;
      }

      /* Try next component */
   }

   if(MLB_MAX_COMPONENTS == i)
   {
      /* Failed: component not found */
      ret = NULL;
   }

   return ret;
}


/* --------------------------------------------------------------------------- loc_its_alive() */
void loc_its_alive(void) {
   /* Write some identity values so that GPP side can verify that DSP is actually running
    *
    *  (note) the first two cachelines (64 bytes) are reserved for GPP<>DSP communication
    *
    */
   sU32 i;
   volatile sU32 *resetBuf = (sU32*) (dsp_config.reset_vector.phys_addr + DSP_RESET_VECTOR_CODESIZE);

   for(i=0; i<16; i++)
   {
      resetBuf[i] = 0;
   }

   for(i=16; i<((DSP_RESET_VECTOR_MINSIZE - DSP_RESET_VECTOR_CODESIZE)/4); i++)
   {
      resetBuf[i] = i;
   }

   BCACHE_wbAll();
}


/* --------------------------------------------------------------------------- mlb_irq__handler() */
void mlb_irq_handler(void *_arg) {
   ////HWI_enter();

   (void)_arg;

   /* Acknowledge interrupt */
   ////DSP_REG(MLB_IRQSTATUS_1) |= (1 << 0); // clear NEWMSGSTATUSUUMB0
   DSP_REG(MLB_IRQSTATUS_1) = (1u << 0); // clear NEWMSGSTATUSUUMB0
   //DSP_REG(MLB_IRQSTATUS_1) = 15;

   // Disable mailbox interrupt to prevent interrupt spam
   //  (workaround for buggy hardware)
   DSP_REG(MLB_IRQENABLE_1) = 0;


   ////DSP_REG(0x86000084u) = num_irqs;
   ////BCACHE_wb((void*)0x86000084u, 4, 1/*wait*/);

   //C64_clearIFR(1 << MLB_INT_VEC_ID);
   //return;


#if 1
   {
      sU32 numMsg = DSP_REG(MLB_MSGSTATUS_0) & 7u; // already reads 0 on DSP-side, works on ARM-side (HW BUG?!!)
      //sU32 numMsg = DSP_REG(MLB_MSGSTATUS_1); //
      ///sU32 numMsg;

      ////if(0 != DSP_REG(MLB_FIFOSTATUS_0))
      if(numMsg > 0)
      {
         sUI fifoIdx = 0;

         ////numMsg = 4;

         while(fifoIdx < numMsg)
         {
            mlb_inbox[fifoIdx] = DSP_REG(MLB_MESSAGE_0);
            //mlb_inbox[fifoIdx] = DSP_REG(MLB_MESSAGE_1);

            // Send reply
            ///DSP_REG(MLB_MESSAGE_1) = 0x900df00du + ((1u + fifoIdx) * 0x100u);
            ////DSP_REG(MLB_MESSAGE_0) = 0x900df00du + (fifoIdx * 0x10u);

            // Next message
            fifoIdx++;
         }

         /* Acknowledge interrupt */
         //DSP_REG(MLB_IRQSTATUS_1) |= (1 << 0); // clear NEWMSGSTATUSUUMB0
         ////DSP_REG(MLB_IRQSTATUS_0) |= (1 << 2); // clear NEWMSGSTATUSUUMB1

         //C64_clearIFR(C64_EINT4);
      }
   }
#endif

   ////mlb_inbox[0] = DSP_REG(MLB_MESSAGE_0);


   //int_event_clear(IVA2_EVT_MAIL_U1_IRQ);
   //CSL_intcEventClear(IVA2_EVT_MAIL_U1_IRQ);

   int_irq_clear(MLB_INT_VEC_ID);
   //CSL_intcInterruptClear(MLB_INT_VEC_ID);

   int_event_clear(IVA2_EVT_MAIL_U1_IRQ);
   C64_clearIFR(1 << MLB_INT_VEC_ID);

   ////DSP_REG(MLB_MESSAGE_1) = 0x900df00d;
   //DSP_REG(MLB_MESSAGE_0) = 0x901df00d;

   //SEM_post(mlb_sem_handle);

   BCACHE_wbInvAll();

   ////HWI_exit();

   num_irqs++;

}


/* --------------------------------------------------------------------------- loc_mlb_reset() */
#if 0
static void loc_mlb_reset(void) {

   DSP_REG(MLB_IRQENABLE_0) = 0;

   DSP_REG(MLB_IRQENABLE_1) = 0;

   DSP_REG(MLB_SYSCONFIG) =
         (DSP_REG(MLB_SYSCONFIG) & ~((1u << 3) | (1u << 4))) |
               (2u << 3) |  // SIDLEMODE
               (1u << 0) |  // AUTOIDLE
               (1u << 1) ;  // SOFTRESET

   while(1 != DSP_REG(MLB_SYSSTATUS))
   {
      /* Wait for RESETDONE bit */
   }

   DSP_REG(MLB_IRQENABLE_0) =
         (
            (1u << 2)  // set NEWMSGENABLEUUMB1 --> Generate IRQ when msg was received from the DSP
          );

   DSP_REG(MLB_IRQENABLE_1) =
         (
            (1u << 0)  // set NEWMSGENABLEUUMB0 --> Generate IRQ when msg was received from the GPP
          );
}
#endif


/* --------------------------------------------------------------------------- mlb_task__entry() */
void mlb_send_to_gpp(sU32 _msg) {
   //loc_mlb_reset();  // workaround for buggy HW that does not clear FIFO after reading

   DSP_REG(MLB_MESSAGE_1) = ~mlb_inbox[0];
}


/* --------------------------------------------------------------------------- mlb_task__entry() */
static void mlb_task__entry(void) {

   num_irqs      = 0;
   num_msg_recvd = 0;

   ///DSP_REG(MLB_MESSAGE_1) = 0x900df00d;

   /* Enable OS timer interrupt */
   //DSP_REG(IVA2_WUGEN_MEVTCLR0) = (1u << 6);
   //C64_clearIFR(1u << 5);//C64_EINT5);
   //C64_enableIER(1u << 5);

   //mlb_sem_handle = SEM_create(0, NULL);

   mlb_debug(0x07000001u);


   /* Check whether MAILBOX registers can be accessed
    *
    *  (note) the "c64.ko" kernel module must have been loaded successfully on the GPP side
    */
   /////if(0 == DSP_REG(CORE_CM_IDLEST1) & 0x80)  // does not work on DSP side, apparently !??!
   {
      sBool bRecvd = 0;
      sU32 iterCount = 0x10000000u;

#if 0
      /* Map MAIL_U1_IVA2_IRQ (55) to mailbox interrupt vector */
      //int_map_event_to_vector(IVA2_EVT_MAIL_U1_IRQ, MLB_INT_VEC_ID);
      //CSL_intcMapEventVector(IVA2_EVT_MAIL_U1_IRQ, MLB_INT_VEC_ID);

      /* Add interrupt handler to HWI dispatcher table */
      //HWI_dispatchPlug(MLB_INT_VEC_ID, (Fxn)mlb_irq_handler, -1, NULL);

      /* Enable event source */
      //int_event_enable(IVA2_EVT_MAIL_U1_IRQ);
      //CSL_intcEventEnable(IVA2_EVT_MAIL_U1_IRQ);

      /* Enable interrupt */
      //int_irq_enable(MLB_INT_VEC_ID);
      //CSL_intcInterruptEnable(MLB_INT_VEC_ID);

#endif


#if 1
      HWI_disable();

      /* Map MAIL_U1_IVA2_IRQ (55) to mailbox interrupt vector (10) */
      HWI_eventMap(
            MLB_INT_VEC_ID,
            IVA2_EVT_MAIL_U1_IRQ
            );

      /* Add interrupt handler to HWI dispatcher table */
      HWI_dispatchPlug(
            MLB_INT_VEC_ID,       // "vecid must be a valid interrupt ID in the range of 0-15."
            (Fxn)mlb_irq_handler,
            -1,                   // dma channel, -1 == use CPU copy
            NULL                  // use default args
            );

      /* Enable event source */
      //int_event_enable(IVA2_EVT_MAIL_U1_IRQ);

      /* Unmask mailbox interrupt (IVA2_IRQ[10]) */
      ////HWI_enableWugen(10);
      DSP_REG(IVA2_WUGEN_MEVTCLR0) = (1u << 10);

      num_irqs = 0x20000000u;

      HWI_enable();

      /* Setup mailbox 0 for receiving messages from the DSP to the GPP, and
       *       mailbox 1 for sending messages to the GPP
       */
      DSP_REG(MLB_IRQENABLE_1) =
            (
               //(1u << 3) |  // Generate IRQ when msg was delivered to the GPP and FIFO slot becomes avail.
               (1u << 0)    // set NEWMSGENABLEUUMB0 --> Generate IRQ when msg was received from the GPP
             );

      /* Enable mailbox interrupt */
      C64_clearIFR(1 << MLB_INT_VEC_ID);//C64_EINT4);
      C64_enableIER(1 << MLB_INT_VEC_ID);//C64_EINT4);
#endif


      // Fill buffer so ARM-side can see that the DSP is doing something
      loc_its_alive();

      mlb_debug(0x07000002u);


      // Send test message to ARM (works)
      //DSP_REG(MLB_MESSAGE_1) = 0x911df11d;

      //DSP_REG(0x86000084u) = DSP_REG(MLB_MSGSTATUS_1) & 7u; // should read 1 but always reads 0
      //BCACHE_wb((void*)0x86000084u, 4, 1/*wait*/);


      //DSP_REG(MLB_MESSAGE_0) = 0x911df11d;
      //for(;;);

      /* Trigger IRQ */
      ////DSP_REG(IVA2_IC_EVTSET_1) = (55 & 31); // does not work

      //for(;;);

      //reset_buf_fill();
      ////DSP_REG(MLB_MESSAGE_1) = 0x911df11d;
      //for(;;);

#if 0
      // Read mailbox revision (0x40) and send to GPP
      {
         sU32 mlbRev = DSP_REG(MLB_REVISION);

         DSP_REG(MLB_MESSAGE_1) = mlbRev;
         //DSP_REG(MLB_MESSAGE_0) = mlbRev;

         //for(;;);
      }
#endif

#if 0
      for(;;)
      {
         DSP_REG(0x86000080u) = ++iterCount;
         BCACHE_wb((void*)0x86000080u, 4, 1/*wait*/);
      }
#endif

      // Send msg to self (--> does not work / triggers no irq)
      ///DSP_REG(MLB_MESSAGE_0) = 0x12345678u;



#if 0
      for(;;)
      {
         DSP_REG(0x86000080u) = ++iterCount;
         BCACHE_wb((void*)0x86000080u, 4, 1/*wait*/);

         /* Poll. fscking interrupts do not work. */
         sU32 numMsg = DSP_REG(MLB_MSGSTATUS_0) & 7u; // already reads 0 (HW BUG??)
         //sU32 numMsg = DSP_REG(MLB_MSGSTATUS_1); //
         ///sU32 numMsg;

         ////if(0 != DSP_REG(MLB_FIFOSTATUS_0))
         if(numMsg > 0)
         {
            sUI fifoIdx = 0;

            //numMsg = 4;

            while(fifoIdx < numMsg)
            {
               mlb_inbox[fifoIdx] = DSP_REG(MLB_MESSAGE_0);
               //mlb_inbox[fifoIdx] = DSP_REG(MLB_MESSAGE_1);

               // Send reply
               DSP_REG(MLB_MESSAGE_1) = 0x900df00du + (fifoIdx * 0x10u);
               //DSP_REG(MLB_MESSAGE_0) = 0x900df00du + (fifoIdx * 0x10u);

               // Next message
               fifoIdx++;
            }

            /* Acknowledge interrupt */
            DSP_REG(MLB_IRQSTATUS_1) |= (1 << 0); // clear NEWMSGSTATUSUUMB0
            //DSP_REG(MLB_IRQSTATUS_0) |= (1 << 2); // clear NEWMSGSTATUSUUMB1

            //C64_clearIFR(C64_EINT4);
         }
      }
#endif

      ////reset_buf_fill();
      ////for(;;);

      // Send test message to GPP
      ////DSP_REG(MLB_MESSAGE_1) = 0x911df11d;

#if 0
      for(;;)
      {
         SEM_pend(mlb_sem_handle, SYS_FOREVER);

         /* Mailbox interrupt occured */

         DSP_REG(0x86000080u) = ++iterCount;
         BCACHE_wb((void*)0x86000080u, 4, 1/*wait*/);


         //if(!bRecvd)
         {
            //if(num_irqs > 0)
            {
               //bRecvd = S_TRUE;

               // Send reply
               //DSP_REG(MLB_MESSAGE_1) = 0x900df00d;
               ////DSP_REG(MLB_MESSAGE_0) = 0x900df00d;
            }
         }
      }
#endif

#if 1


#if 0
#if 1
      // Trigger INT10 interrupt (works)
      asm(" mvkl 0x400, a0");
      asm(" mvc a0, ISR");
#endif

#if 1
      // Trigger INT10 interrupt (works)
      asm(" mvkl 0x400, a0");
      asm(" mvc a0, ISR");
#endif

#if 1
      // Trigger INT10 interrupt (works)
      asm(" mvkl 0x400, a0");
      asm(" mvc a0, ISR");
#endif

#if 1
      // Trigger INT10 interrupt (works)
      asm(" mvkl 0x400, a0");
      asm(" mvc a0, ISR");
#endif
#endif

      {
         sUI oldNumIrqs = num_irqs;
         sUI oldMsgSerial = 0;

         for(;;)
         {
            if(num_irqs != oldNumIrqs)
            {
               sUI newMsgSerial;

               mlb_debug(0x07000003u);

               oldNumIrqs = num_irqs;

#ifdef PANDORA_VID2_MSG_HACK
               /* should not be necessary, this is a register!?! */
               BCACHE_inv((Ptr)DSP_MLB_RAM_MSGSERIAL, 64, 1/*wait*/);
#else
               BCACHE_inv((Ptr)DSP_MLB_RAM_MSGSERIAL, 32, 1/*wait*/);
#endif /* PANDORA_VID2_MSG_HACK */

               newMsgSerial = DSP_REG(DSP_MLB_RAM_MSGSERIAL);

               DSP_REG(DSP_MLB_RAM_DSPIRQCOUNT) = num_irqs;

               DSP_REG(DSP_MLB_RAM_DSPLASTREADMSG) = newMsgSerial;

               mlb_debug(0x07000004u);

               BCACHE_wbInv((void*)DSP_MLB_RAM_DSPIRQCOUNT, 8, 1/*wait*/);

               if(oldMsgSerial != newMsgSerial)
               {
                  sU32 rpcDebug = 0;

                  oldMsgSerial = newMsgSerial;

                  //DSP_REG(DSP_MLB_RAM_DSPLASTREADMSG) = mlb_inbox[0];
                  //BCACHE_wb((void*)DSP_MLB_RAM_DSPLASTREADMSG, 4, 1/*wait*/);

                  mlb_debug(0x07000005u);

                  /* Try to execute requested command */
#if 1
                  {
                     const dsp_component_t *comp = loc_component_find_by_id(
#ifdef PANDORA_VID2_MSG_HACK
                           (dsp_component_id_t) (DSP_REG(DSP_MLB_RAM_MSGTODSP_COMP) & 0xFFFFu)
#else
                           *(dsp_component_id_t*)(DSP_MLB_RAM_MSGTODSP + 0u)
#endif /* PANDORA_VID2_MSG_HACK */
                           );

                     if(NULL != comp)
                     {
                        sU32 err;

                        /* Execute component command
                         *
                         *  (note) the first +4u skips the "client" field of the dsp_msg_t structure
                         *  (note) +2u then addresses the cmd_id (sU16)
                         *  (note) the remaining +4u resp. +8u address the arg1 resp. arg2 fields
                         *
                         *  (note) also see "msg.h"
                         */
#ifdef PANDORA_VID2_MSG_HACK
                        sU32 ret1;
                        sU32 ret2;
#endif /* PANDORA_VID2_MSG_HACK */

                        err = comp->fxns.exec(
#ifdef PANDORA_VID2_MSG_HACK
                              (dsp_component_cmd_t) (DSP_REG(DSP_MLB_RAM_MSGTODSP_COMP) >> 16),
                              DSP_REG(DSP_MLB_RAM_MSGTODSP_ARG1),
                              DSP_REG(DSP_MLB_RAM_MSGTODSP_ARG2),
                              &ret1,
                              &ret2
#else
                              *(dsp_component_cmd_t*)(DSP_MLB_RAM_MSGTODSP + 4u + 2u),
                              DSP_REG(DSP_MLB_RAM_MSGTODSP + 4u + 4u),
                              DSP_REG(DSP_MLB_RAM_MSGTODSP + 4u + 8u),
                              (volatile sU32*) (DSP_MLB_RAM_MSGTOGPP + 4u + 4u),
                              (volatile sU32*) (DSP_MLB_RAM_MSGTOGPP + 4u + 8u)
#endif /* PANDORA_VID2_MSG_HACK */
                              );

#ifdef PANDORA_VID2_MSG_HACK
                        DSP_REG(DSP_MLB_RAM_MSGTOGPP_RET1) = ret1;
                        DSP_REG(DSP_MLB_RAM_MSGTOGPP_RET2) = ret2;

                        /* should not be necessary, this is a register !?!?
                         *  (-> does not make any difference)
                         */
                        BCACHE_wbInv((void*)DSP_MLB_RAM_MSGSERIAL, 64, 1/*wait*/);
#else
                        BCACHE_wbInv((void*)DSP_MLB_RAM_MSGTOGPP, 32, 1/*wait*/);
#endif /* PANDORA_VID2_MSG_HACK */

                        if(0 == err)
                        {
                           /* Ok, command completed successfully */
                           rpcDebug = 0x70000000u;

#ifdef PANDORA_VID2_MSG_HACK
                           rpcDebug |= ((DSP_REG(DSP_MLB_RAM_MSGTOGPP_RET1)) & 255) << 8u;
#else
                           rpcDebug |= ((DSP_REG(DSP_MLB_RAM_MSGTOGPP + 4u + 4u)) & 255) << 8u;
#endif /* PANDORA_VID2_MSG_HACK */
                        }
                        else
                        {
                           /* (todo) propagate error code to GPP */
                           rpcDebug = 0x30000000u;

                           rpcDebug |=
                                 ((err & 255u) << 16u) |
#ifdef PANDORA_VID2_MSG_HACK
                                 ((DSP_MLB_RAM_MSGTODSP_COMP >> 16) << 8u)
#else
                                 ((*(dsp_component_cmd_t*)(DSP_MLB_RAM_MSGTODSP + 4u + 2u)) << 8u)
#endif /* PANDORA_VID2_MSG_HACK */
                                 ;
                        }

                     }
                     else
                     {
                        /* Failed: component not found */
                        rpcDebug = 0x10000000u;
                     }
                  }
#endif


                  /* Write back caches so GPP can see the data */
                  BCACHE_wbInvAll();

                  mlb_debug(0x07000006u | rpcDebug);

                  /* Send reply to trigger IRQ on GPP-side.
                   *
                   *  (note) the content of the message register does not matter,
                   *          it's only written to in order to trigger an interrupt.
                   *
                   *  (note) even if the command failed or the requested component does not exist,
                   *          a reply message must be sent.
                   */
                  mlb_send_to_gpp(~mlb_inbox[0]);

                  mlb_debug(0x07000007u | rpcDebug);

                  iterCount = 0;

               } // if(oldMsgSerial != newMsgSerial)

            } // if(num_irqs != oldNumIrqs)

            iterCount++;

            if(0 == (iterCount & 1023))
            {
               DSP_REG(DSP_MLB_RAM_WATCHDOG) = iterCount;
               BCACHE_wbInv((void*)DSP_MLB_RAM_WATCHDOG, 4, 1/*wait*/);
            }

            ////TSK_sleep(1); // never returns

         } // for(;;);
      }
#endif
   }
}


/* --------------------------------------------------------------------------- watchdog_task__entry() */
#if 0
static void watchdog_task__entry(void) {
   sUI count = 0;

   for(;;)
   {
      DSP_REG(0x86000088u) = ++count;
      BCACHE_wb((void*)0x86000088u, 4, 1/*wait*/);

      TSK_sleep(10); // never returns
   }
}
#endif


/* --------------------------------------------------------------------------- mlb_init() */
void mlb_init(void) {
   /* (note) must be called in main() before registering components */

   memset(components, 0, sizeof(components));

   num_components = 0;
}


/* --------------------------------------------------------------------------- mlb_start() */
sBool mlb_component_register (dsp_component_id_t     _compId,
                              const dsp_component_t *_comp
                              ) {
   sBool ret;

   if(_compId > 0)
   {
      if(NULL != _comp)
      {
         if(num_components < MLB_MAX_COMPONENTS)
         {
            if(NULL != _comp->fxns.exec)
            {
               components[num_components] = *_comp;

               components[num_components].component_id = _compId;

               num_components++;

               /* Succeeded */
               ret = S_TRUE;
            }
            else
            {
               /* Failed: no exec function */
               ret = S_FALSE;
            }
         }
         else
         {
            /* Failed: maximum number of components exceeded */
            ret = S_FALSE;
         }
      }
      else
      {
         /* Failed: component reference is NULL */
         ret = S_FALSE;
      }
   }
   else
   {
      /* Failed: illegal component id */
      ret = S_FALSE;
   }

   return ret;
}


/* --------------------------------------------------------------------------- mlb_start() */
void mlb_start(void) {

   ////DSP_REG(MLB_MESSAGE_1) = 0x900df00d;
   ////reset_buf_fill();
   ////for(;;);

   //watchdog_task_attrs = TSK_ATTRS;
   //watchdog_task_attrs.name     = "wdog";
   //watchdog_task_attrs.priority = TSK_MINPRI;

   //watchdog_task_handle = TSK_create((Fxn)watchdog_task__entry, &watchdog_task_attrs);

   /* Initialize components */
   {
      sUI i;

      for(i=0; i<num_components; i++)
      {
         if(NULL != components[i].fxns.init)
         {
            if(!components[i].fxns.init())
            {
               /* Failed: component failed to initialize */
               mlb_panic();
            }
         }

         /* Next component */
      }
   }

   /* Create ARM<>DSP mailbox task
    *
    *  (todo) use tconf to create the task statically (size opt.)
    */
   mlb_task_attrs = TSK_ATTRS;
   mlb_task_attrs.name     = "mlb";
   mlb_task_attrs.priority = TSK_MINPRI;

   mlb_task_handle = TSK_create((Fxn)mlb_task__entry, &mlb_task_attrs);
}


/* --------------------------------------------------------------------------- mlb_debug() */
void mlb_debug(sU32 _val) {

   DSP_REG(DSP_MLB_RAM_DSPDEBUG) = _val;

   BCACHE_wbInv((void*)DSP_MLB_RAM_DSPDEBUG, 4, 1/*wait*/);
}


/* --------------------------------------------------------------------------- mlb_panic() */
void mlb_panic(void) {

   /* (todo) signal GPP-side that it needs to reboot the DSP */

   for(;;)
   {
      /* Loop a while.. loop..forever! */

      /* (todo) execute idle instruction ?! */
   }
}
