/* ----
 * ---- file   : kmod.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : C64 kernel module. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 06Sep2013, 08Sep2013, 10Sep2013, 11Sep2013, 14Sep2013, 15Sep2013
 * ----
 * ----
 */

#undef __KERNEL__
#define __KERNEL__

#undef MODULE
#define MODULE defined

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/export.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#include "../include/types.h"
#include "../include/dsp_common.h"
#include "../include/msg.h"

#include "kmod.h"
#include "reg.h"
#include "mlb.h"
#include "dsp_c64.h"
#include "dev.h"


/*--------------------------------------------------------------------------- global vars */
sUI c64_debug_level = 0;


/*--------------------------------------------------------------------------- module vars */


/*--------------------------------------------------------------------------- c64_init() */
static int __init c64_init(void) {
   int ret;

   if(c64_debug_level > 0)
   {
      printk(KERN_INFO "c64::c64_init: ENTER\n");
   }

   /* Put DSP into reset and disable MMU */
   dsp_init();

   /* Setup mailbox and interrupt handler */
   if(mlb_init())
   {
      /* Register "/dev/c64" char device */
      if(dev_init())
      {
         /* Succeeded */
         ret = 0;
      }
      else
      {
         /* Failed: dev_init() */
         mlb_exit();
         dsp_exit();

         ret = -1;
      }
   }
   else
   {
      /* Failed: mlb_init() */

      dsp_exit();

      ret = -1;
   }

   if(c64_debug_level > 0)
   {
      printk(KERN_INFO "c64::c64_init: LEAVE ret=%d\n", ret);
   }
   else
   {
      printk(KERN_INFO "c64: init ok. module build date: %s\n", __DATE__
#ifdef PANDORA_VID2_MSG_HACK
             "  (**VID2_MSG_HACK**)"
#endif /* PANDORA_VID2_MSG_HACK */
             );
   }

   return ret;
}


/*--------------------------------------------------------------------------- c64_exit() */
static void __exit c64_exit(void) {

   if(c64_debug_level > 0)
   {
      printk(KERN_INFO "c64::c64_exit: ENTER\n");
   }

   /* Unregister char device */
   dev_exit();

   /* Disable mailbox interrupt and unregister handler */
   mlb_exit();

   /* Power off IVA2 subsystem */
   dsp_exit();

   if(c64_debug_level >= 10)
   {
      printk(KERN_INFO "c64::c64_exit: LEAVE\n");
   }
}


/*--------------------------------------------------------------------------- c64_param_set_debug_level() */
static int c64_param_set_debug_level(const char *_val, struct kernel_param *_kp) {
   (void)_kp;

   sscanf(_val, "%u", &c64_debug_level);

   return 0;
}


/*--------------------------------------------------------------------------- module exports */
MODULE_AUTHOR("Bastian Spiegel");
MODULE_LICENSE("GPL");

MODULE_DESCRIPTION("TMS320C6000 init + ARM<>DSP messaging driver");
MODULE_SUPPORTED_DEVICE("omap3530, dm3730 (Open Pandora)");

////module_param(c64_debug_level, int, 0);
module_param_call(debug, &c64_param_set_debug_level, NULL, NULL, 0200);
MODULE_PARM_DESC(debug, "debug level (0=no debug output, 20=trace reg access)");

module_init(c64_init);
module_exit(c64_exit);
