/* ----
 * ---- file   : dsp_common.h
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the LESSER GNU GENERAL PUBLIC LICENSE (LGPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#LGPL or COPYING_LGPL for further information.
 * ----
 * ---- info   : Common DSP types and structs. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 06Sep2013, 08Sep2013, 11Sep2013, 13Sep2013, 14Sep2013, 15Sep2013
 * ----
 * ----
 */

#ifndef __C64_TOOLS_DSP_COMMON_H__
#define __C64_TOOLS_DSP_COMMON_H__

#include "cplusplus_begin.h"

/* If defined, enable hack to (hopefully) get rid of cache coherency issues.
 *
 *  The hack abuses the DISPC_VID2_FIR_COEF_HI registers (0x48050580 +0, +8, +16, +24)
 *   for GPP<>DSP IPC.
 *
 *  Yes, that's not a solution I'm particularly proud of and if anyone knows how to
 *   do this properly, let me know! (resp. code it and test it to make sure it rly works!)
 *
 */
//#define PANDORA_VID2_MSG_HACK defined


#define DSP_RESET_VECTOR_CODESIZE  (0x00000080u)  // size of RESET_VECTOR memory region used for boot code
                                                  //  (actually only 32 bytes are used)
#define DSP_RESET_VECTOR_MINSIZE   (0x00001000u)  // minimum size of RESET_VECTOR memory region

#define DSP_RESET_VECTOR_MSGSIZE   (DSP_RESET_VECTOR_MINSIZE - DSP_RESET_VECTOR_CODESIZE)


#define DSP_PAGESIZE (4096u)  // (todo) use sysconf(_SC_PAGE_SIZE) [linux]

#define DSP_PAGESIZE_ROUND(a) ( ( ((a)+(DSP_PAGESIZE-1)) / DSP_PAGESIZE ) * DSP_PAGESIZE )

#define DSP_PAGESIZE_ALIGN(a) ( ( (a) / DSP_PAGESIZE ) * DSP_PAGESIZE )


/* for IPC / from GPP to DSP */
#ifdef PANDORA_VID2_MSG_HACK

#define DSP_HACK_BASE (0x48050580u)

#define DSP_MLB_RAM_MSGSERIAL       (DSP_HACK_BASE +  0u)

#define DSP_MLB_RAM_MSGTODSP_COMP   (DSP_HACK_BASE +  8u)  // component id (lo) / command id (hi)
#define DSP_MLB_RAM_MSGTODSP_ARG1   (DSP_HACK_BASE + 16u)
#define DSP_MLB_RAM_MSGTODSP_ARG2   (DSP_HACK_BASE + 24u)

#else

#define DSP_MLB_RAM_MSGSERIAL       (dsp_config.reset_vector.phys_addr + DSP_RESET_VECTOR_CODESIZE +  0u)  // 0x86000080

#endif /* PANDORA_VID2_MSG_HACK */

/* (note) when the hack is enabled, this memory area simply stores the client handle (not read by the DSP, anyway) */
#define DSP_MLB_RAM_MSGTODSP        (dsp_config.reset_vector.phys_addr + DSP_RESET_VECTOR_CODESIZE + 16u)  // 0x86000090


/* for IPC / from DSP to GPP */
#ifdef PANDORA_VID2_MSG_HACK

#define DSP_MLB_RAM_MSGTOGPP_RET1   (DSP_HACK_BASE + 32u)
#define DSP_MLB_RAM_MSGTOGPP_RET2   (DSP_HACK_BASE + 40u)

#else

#define DSP_MLB_RAM_MSGTOGPP        (dsp_config.reset_vector.phys_addr + DSP_RESET_VECTOR_CODESIZE + 32u)  // 0x860000A0

#endif /* PANDORA_VID2_MSG_HACK */

#define DSP_MLB_RAM_WATCHDOG        (dsp_config.reset_vector.phys_addr + DSP_RESET_VECTOR_CODESIZE + 48u)  // 0x860000B0  (debugging)
#define DSP_MLB_RAM_DSPIRQCOUNT     (dsp_config.reset_vector.phys_addr + DSP_RESET_VECTOR_CODESIZE + 52u)  // 0x860000B4  (debugging)
#define DSP_MLB_RAM_DSPLASTREADMSG  (dsp_config.reset_vector.phys_addr + DSP_RESET_VECTOR_CODESIZE + 56u)  // 0x860000B8  (debugging)
#define DSP_MLB_RAM_DSPDEBUG        (dsp_config.reset_vector.phys_addr + DSP_RESET_VECTOR_CODESIZE + 60u)  // 0x860000BC


typedef struct {
   sU32 phys_addr;  // on DSP-side
   sU32 virt_addr;  // on ARM-side (mmap'd) (!= 0 while mapped)
   sU32 size;       // number of bytes
} dsp_mem_region_t;


#include "cplusplus_end.h"

#endif /* __C64_TOOLS_DSP_COMMON_H__ */
