/* ----
 * ---- file   : osal_mmap.c
 * ---- author : Bastian Spiegel <bs@tkscript.de>
 * ---- legal  : (c) 2013 by Bastian Spiegel. 
 * ----          Distributed under terms of the GNU GENERAL PUBLIC LICENSE (GPL). See 
 * ----          http://www.gnu.org/licenses/licenses.html#GPL or COPYING for further information.
 * ----
 * ---- info   : Linux OS abstraction layer. This is part of the "c64_tools" package.
 * ----
 * ---- changed: 06Sep2013, 08Sep2013
 * ----
 * ----
 */

#include <stdio.h>
#include <stdlib.h>
//#include <string.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>

#include "../../../include/types.h"
#include "../../../include/log.h"
#include "../../../include/osal_err.h"
#include "../../../include/osal.h"


#ifndef HAVE_OS_LINUX
#error osal_linux.c must be compiled for Linux (no cygwin support here)
#endif


/*--------------------------------------------------------------------------- module vars */
static int fd_mem;


/*--------------------------------------------------------------------------- osal_init() */
osal_error_t S_CALL osal_init(void) {
   osal_error_t ret;

   fd_mem = open("/dev/mem", O_RDWR | O_SYNC);

   if(-1 != fd_mem)
   {
      /* Succeeded */
      ret = OSAL_ERR_OK;
   }
   else
   {
      // Failed: open()
      log_printf(LOG_ERROR "osal_init: open(\"/dev/mem\") failed (errno=%d). Are you root?\n", errno);

      ret = OSAL_ERR_MEM_OPEN;
   }

   return ret;
}


/*--------------------------------------------------------------------------- osal_exit() */
void S_CALL osal_exit(void) {
   if(-1 != fd_mem)
   {
      close(fd_mem);
   }
}


/*--------------------------------------------------------------------------- osal_alloc() */
void * S_CALL osal_alloc(sSize _numBytes) {
   void *ret;

   ret = malloc(_numBytes);

   if(NULL == ret)
   {
      log_printf(LOG_ERROR "osal_alloc: failed to allocate %u bytes.\n", _numBytes);
   }
   
   return ret;
}


/*--------------------------------------------------------------------------- osal_free() */
void S_CALL osal_free(void *_ptr) {
   if(NULL != _ptr)
   {
      free(_ptr);
   }
}


/*--------------------------------------------------------------------------- osal_mmap() */
void * S_CALL osal_mmap(void *_physAddr, sSize _numBytes) {
   void *ret;

   ret = mmap(NULL /*addr*/,
              _numBytes,
              PROT_READ | PROT_WRITE,
              MAP_SHARED,
              fd_mem,
              (off_t) _physAddr
              );

   if(MAP_FAILED == ret)
   {
      log_printf(LOG_ERROR "osal_mmap: mmap() failed (errno=%d)\n", errno);

      ret = NULL;
   }

   return ret;
}


/*--------------------------------------------------------------------------- osal_munmap() */
osal_error_t S_CALL osal_munmap(void *_virtAddr, sSize _numBytes) {
   osal_error_t ret;

   int r = munmap(_virtAddr, _numBytes);
      
   if(0 == r)
   {
      // Succeeded
      ret = OSAL_ERR_OK;
   }
   else
   {
      // Failed: munmap()
      ret = OSAL_ERR_MEM_UNMAP;
   }

   return ret;
}
